import functools
import inspect
from typing import Any

import dobles.lifecycle
from dobles.call_count_accumulator import CallCountAccumulator
from dobles.exceptions import MockExpectationError, VerifyingBuiltinDoubleArgumentError
from dobles.verification import verify_arguments

_any = object()


async def _async_return_value(*args, **kwargs):
    return None


def _maybe_async(is_async: bool, return_value: Any):
    if is_async:
        return make_it_async(return_value)
    return return_value


def make_it_async(
    return_value=None,
    exception=None,
):
    async def inner(*args, **kwargs):
        if exception:
            raise exception
        return return_value

    return inner()


def verify_count_is_non_negative(func):
    @functools.wraps(func)
    def inner(self, arg):
        if arg < 0:
            raise TypeError(func.__name__ + " requires one positive integer argument")
        return func(self, arg)

    return inner


def check_func_takes_args(func):
    arg_spec = inspect.getfullargspec(func)
    return any([arg_spec.args, arg_spec.varargs, arg_spec.varkw, arg_spec.defaults])


def build_argument_repr_string(args, kwargs):
    args = [repr(x) for x in args]
    kwargs = ["{}={!r}".format(k, v) for k, v in kwargs.items()]
    return "({})".format(", ".join(args + kwargs))


class Allowance(object):
    """An individual method allowance (stub)."""

    def __init__(self, target, method_name, caller):
        """
        :param Target target: The object owning the method to stub.
        :param str method_name: The name of the method to stub.
        """

        self._target = target
        self._method_name = method_name
        self._caller = caller
        self.args = _any
        self.kwargs = _any
        self._custom_matcher = None
        self._is_satisfied = True
        self._call_counter = CallCountAccumulator()

        self.is_async: bool = target.is_attr_async(method_name)
        if self.is_async:
            self._return_value = _async_return_value
        else:
            self._return_value = lambda *args, **kwargs: None

    def and_raise(self, exception, *args, **kwargs):
        """Causes the double to raise the provided exception when called.

        If provided, additional arguments (positional and keyword) passed to
        `and_raise` are used in the exception instantiation.

        :param Exception exception: The exception to raise.
        """

        def proxy_exception(*proxy_args, **proxy_kwargs):
            raise exception

        async def async_proxy_exception(*proxy_args, **proxy_kwargs):
            raise exception

        self._return_value = async_proxy_exception if self.is_async else proxy_exception
        return self

    def and_return(self, *return_values):
        """Set a return value for an allowance

        Causes the double to return the provided values in order.  If multiple
        values are provided, they are returned one at a time in sequence as the double is called.
        If the double is called more times than there are return values, it should continue to
        return the last value in the list.


        :param object return_values: The values the double will return when called,
        """

        if not return_values:
            raise TypeError("and_return() expected at least 1 return value")

        return_values = list(return_values)
        final_value = return_values.pop()

        self.and_return_result_of(
            lambda: return_values.pop(0) if return_values else final_value
        )
        return self

    def and_return_result_of(self, return_value):
        """Causes the double to return the result of calling the provided value.

        :param return_value: A callable that will be invoked to determine the double's return value.
        :type return_value: any callable object
        """

        if not check_func_takes_args(return_value):
            self._return_value = lambda *args, **kwargs: _maybe_async(
                self.is_async,
                return_value(),
            )
        else:
            self._return_value = _maybe_async(self.is_async, return_value)

        return self

    def is_satisfied(self):
        """Returns a boolean indicating whether or not the double has been satisfied.

        Stubs are always satisfied, but mocks are only satisfied if they've been
        called as was declared.

        :return: Whether or not the double is satisfied.
        :rtype: bool
        """
        return self._is_satisfied

    def with_args(self, *args, **kwargs):
        """Declares that the double can only be called with the provided arguments.

        :param args: Any positional arguments required for invocation.
        :param kwargs: Any keyword arguments required for invocation.
        """

        self.args = args
        self.kwargs = kwargs
        self.verify_arguments()
        return self

    def with_args_validator(self, matching_function):
        """Define a custom function for testing arguments

        :param func matching_function:  The function used to test arguments passed to the stub.
        """
        self.args = None
        self.kwargs = None
        self._custom_matcher = matching_function
        return self

    def __call__(self, *args, **kwargs):
        """A short hand syntax for with_args

        Allows callers to do:
            allow(module).foo.with_args(1, 2)
        With:
            allow(module).foo(1, 2)

        :param args: Any positional arguments required for invocation.
        :param kwargs: Any keyword arguments required for invocation.
        """
        return self.with_args(*args, **kwargs)

    def with_no_args(self):
        """Declares that the double can only be called with no arguments."""

        self.args = ()
        self.kwargs = {}
        self.verify_arguments()
        return self

    def satisfy_any_args_match(self):
        """Returns a boolean indicating whether or not the stub will accept arbitrary arguments.

        This will be true unless the user has specified otherwise using ``with_args`` or
        ``with_no_args``.

        :return: Whether or not the stub accepts arbitrary arguments.
        :rtype: bool
        """

        return self.args is _any and self.kwargs is _any

    def satisfy_exact_match(self, args, kwargs):
        """Returns a boolean indicating whether or not the stub will accept the provided arguments.

        :return: Whether or not the stub accepts the provided arguments.
        :rtype: bool
        """

        if self.args is None and self.kwargs is None:
            return False
        elif self.args is _any and self.kwargs is _any:
            return True
        elif args == self.args and kwargs == self.kwargs:
            return True
        elif len(args) != len(self.args) or len(kwargs) != len(self.kwargs):
            return False

        if not all(x == y or y == x for x, y in zip(args, self.args)):
            return False

        for key, value in self.kwargs.items():
            if key not in kwargs:
                return False
            elif not (kwargs[key] == value or value == kwargs[key]):
                return False

        return True

    def satisfy_custom_matcher(self, args, kwargs):
        """Return a boolean indicating if the args satisfy the stub

        :return: Whether or not the stub accepts the provided arguments.
        :rtype: bool
        """
        if not self._custom_matcher:
            return False
        try:
            return self._custom_matcher(*args, **kwargs)
        except Exception:
            return False

    def return_value(self, *args, **kwargs):
        """Extracts the real value to be returned from the wrapping callable.

        :return: The value the double should return when called.
        """

        self._called()
        return self._return_value(*args, **kwargs)

    def verify_arguments(self, args=None, kwargs=None):
        """Ensures that the arguments specified match the signature of the real method.

        :raise: ``VerifyingDoubleError`` if the arguments do not match.
        """

        args = self.args if args is None else args
        kwargs = self.kwargs if kwargs is None else kwargs

        try:
            verify_arguments(self._target, self._method_name, args, kwargs)
        except VerifyingBuiltinDoubleArgumentError:
            if dobles.lifecycle.ignore_builtin_verification():
                raise

    @verify_count_is_non_negative
    def exactly(self, n):
        """Set an exact call count allowance

        :param integer n:
        """

        self._call_counter.set_exact(n)
        return self

    @verify_count_is_non_negative
    def at_least(self, n):
        """Set a minimum call count allowance

        :param integer n:
        """

        self._call_counter.set_minimum(n)
        return self

    @verify_count_is_non_negative
    def at_most(self, n):
        """Set a maximum call count allowance

        :param integer n:
        """

        self._call_counter.set_maximum(n)
        return self

    def never(self):
        """Set an expected call count allowance of 0"""

        self.exactly(0)
        return self

    def once(self):
        """Set an expected call count allowance of 1"""

        self.exactly(1)
        return self

    def twice(self):
        """Set an expected call count allowance of 2"""

        self.exactly(2)
        return self

    @property
    def times(self):
        return self

    time = times

    def _called(self):
        """Indicate that the allowance was called

        :raise MockExpectationError if the allowance has been called too many times
        """

        if self._call_counter.called().has_too_many_calls():
            self.raise_failure_exception()

    def raise_failure_exception(self, expect_or_allow="Allowed"):
        """Raises a ``MockExpectationError`` with a useful message.

        :raise: ``MockExpectationError``
        """

        raise MockExpectationError(
            "{} '{}' to be called {}on {!r} with {}, but was not. ({}:{})".format(
                expect_or_allow,
                self._method_name,
                self._call_counter.error_string(),
                self._target.obj,
                self._expected_argument_string(),
                self._caller.filename,
                self._caller.lineno,
            )
        )

    def _expected_argument_string(self):
        """Generates a string describing what arguments the double expected.

        :return: A string describing expected arguments.
        :rtype: str
        """

        if self.args is _any and self.kwargs is _any:
            return "any args"
        elif self._custom_matcher:
            return "custom matcher: '{}'".format(self._custom_matcher.__name__)
        else:
            return build_argument_repr_string(self.args, self.kwargs)



from inspect import isclass

from dobles.exceptions import VerifyingDoubleImportError
from dobles.object_double import ObjectDouble
from dobles.utils import get_module, get_path_components


def _get_dobles_target(module, class_name, path):
    """Validate and return the class to be doubled.

    :param module module: The module that contains the class that will be doubled.
    :param str class_name: The name of the class that will be doubled.
    :param str path: The full path to the class that will be doubled.
    :return: The class that will be doubled.
    :rtype: type
    :raise: ``VerifyingDoubleImportError`` if the target object doesn't exist or isn't a class.
    """

    try:
        dobles_target = getattr(module, class_name)
        if isinstance(dobles_target, ObjectDouble):
            return dobles_target._dobles_target

        if not isclass(dobles_target):
            raise VerifyingDoubleImportError(
                "Path does not point to a class: {}.".format(path)
            )

        return dobles_target
    except AttributeError:
        raise VerifyingDoubleImportError("No object at path: {}.".format(path))


class InstanceDouble(ObjectDouble):
    """A pure double representing an instance of the target class.

    Any kwargs supplied will be set as attributes on the instance that is
    created.

    ::

        user = InstanceDouble('myapp.User', name='Bob Barker')

    :param str path: The absolute module path to the class.
    """

    def __init__(self, path, **kwargs):
        module_path, class_name = get_path_components(path)
        module = get_module(module_path, path)
        self._dobles_target = _get_dobles_target(module, class_name, path)
        for k, v in kwargs.items():
            setattr(self, k, v)



def pluralize(word, count):
    return word if count == 1 else word + "s"


class CallCountAccumulator(object):
    def __init__(self):
        self._call_count = 0

    def set_exact(self, n):
        """Set an exact call count expectation

        :param integer n:
        """

        self._exact = n

    def set_minimum(self, n):
        """Set a minimum call count expectation

        :param integer n:
        """

        self._minimum = n

    def set_maximum(self, n):
        """Set a maximum call count expectation

        :param integer n:
        """

        self._maximum = n

    def has_too_many_calls(self):
        """Test if there have been too many calls

        :rtype boolean
        """

        if self.has_exact and self._call_count > self._exact:
            return True
        if self.has_maximum and self._call_count > self._maximum:
            return True
        return False

    def has_too_few_calls(self):
        """Test if there have not been enough calls

        :rtype boolean
        """

        if self.has_exact and self._call_count < self._exact:
            return True
        if self.has_minimum and self._call_count < self._minimum:
            return True
        return False

    def has_incorrect_call_count(self):
        """Test if there have not been a valid number of calls

        :rtype boolean
        """

        return self.has_too_few_calls() or self.has_too_many_calls()

    def has_correct_call_count(self):
        """Test if there have been a valid number of calls

        :rtype boolean
        """

        return not self.has_incorrect_call_count()

    def never(self):
        """Test if the number of expect is 0

        :rtype: boolean
        """

        return self.has_exact and self._exact == 0

    def called(self):
        """Increment the call count"""

        self._call_count += 1
        return self

    @property
    def count(self):
        """Extract the current call count

        :rtype integer
        """

        return self._call_count

    @property
    def has_minimum(self):
        """Test if self has a minimum call count set

        :rtype boolean
        """

        return getattr(self, "_minimum", None) is not None

    @property
    def has_maximum(self):
        """Test if self has a maximum call count set

        :rtype boolean
        """

        return getattr(self, "_maximum", None) is not None

    @property
    def has_exact(self):
        """Test if self has an exact call count set

        :rtype boolean
        """

        return getattr(self, "_exact", None) is not None

    def _restriction_string(self):
        """Get a string explaining the expectation currently set

        e.g `at least 5 times`, `at most 1 time`, or `2 times`

        :rtype string
        """

        if self.has_minimum:
            string = "at least "
            value = self._minimum
        elif self.has_maximum:
            string = "at most "
            value = self._maximum
        elif self.has_exact:
            string = ""
            value = self._exact

        return (string + "{} {}").format(value, pluralize("time", value))

    def error_string(self):
        """Returns a well formed error message

        e.g at least 5 times but was called 4 times

        :rtype string
        """

        if self.has_correct_call_count():
            return ""

        return "{} instead of {} {} ".format(
            self._restriction_string(), self.count, pluralize("time", self.count)
        )



from dobles.exceptions import UnallowedMethodCallError
from dobles.instance_double import InstanceDouble
from dobles.target import Target
from dobles.verification import verify_arguments


def patch_class(input_class):
    """Create a new class based on the input_class.

    :param class input_class:  The class to patch.
    :rtype class:
    """

    class Instantiator(object):
        @classmethod
        def _dobles__new__(self, *args, **kwargs):
            pass

    new_class = type(input_class.__name__, (input_class, Instantiator), {})

    return new_class


class ClassDouble(InstanceDouble):
    """
    A pure double representing the target class.

    ::

        User = ClassDouble('myapp.User')

    :param str path: The absolute module path to the class.
    """

    is_class = True

    def __init__(self, path):
        super(ClassDouble, self).__init__(path)
        self._dobles_target = patch_class(self._dobles_target)
        self._target = Target(self._dobles_target)

    def __call__(self, *args, **kwargs):
        """Verify arguments and proxy to _dobles__new__

        :rtype obj:
        :raises VerifyingDoubleArgumentError: If args/kwargs don't match the expected arguments of
            __init__ of the underlying class.
        """
        verify_arguments(self._target, "_dobles__new__", args, kwargs)
        return self._dobles__new__(*args, **kwargs)

    def _dobles__new__(self, *args, **kwargs):
        """Raises an UnallowedMethodCallError

        NOTE: This method is here only to raise if it has not been stubbed
        """
        raise UnallowedMethodCallError(
            "Cannot call __new__ on a ClassDouble without stubbing it"
        )



from dobles.allowance import Allowance
from dobles.expectation import Expectation
from dobles.proxy_method import ProxyMethod
from dobles.verification import verify_method


class MethodDouble(object):
    """A double of an individual method."""

    def __init__(self, method_name, target):
        """
        :param str method_name: The name of the method to double.
        :param Target target: A ``Target`` object containing the object with the method to double.
        """

        self._method_name = method_name
        self._target = target

        self._verify_method()

        self._allowances = []
        self._expectations = []

        self._proxy_method = ProxyMethod(
            target,
            method_name,
            lambda args, kwargs: self._find_matching_double(args, kwargs),
        )

    def add_allowance(self, caller):
        """Adds a new allowance for the method.

        :param: tuple caller: A tuple indicating where the method was called
        :return: The new ``Allowance``.
        :rtype: Allowance
        """

        allowance = Allowance(self._target, self._method_name, caller)
        self._allowances.insert(0, allowance)
        return allowance

    def add_expectation(self, caller):
        """Adds a new expectation for the method.

        :return: The new ``Expectation``.
        :rtype: Expectation
        """

        expectation = Expectation(self._target, self._method_name, caller)
        self._expectations.insert(0, expectation)
        return expectation

    def restore_original_method(self):
        """Removes the proxy method on the target and replaces it with its original value."""

        self._proxy_method.restore_original_method()

    def verify(self):
        """Verifies all expectations on the method.

        :raise: ``MockExpectationError`` on the first expectation that is not satisfied, if any.
        """

        for expectation in self._expectations:
            if not expectation.is_satisfied():
                expectation.raise_failure_exception()

    def _find_matching_allowance(self, args, kwargs):
        """Return a matching allowance.

        Returns the first allowance that matches the ones declared. Tries one with specific
        arguments first, then falls back to an allowance that allows arbitrary arguments.

        :return: The matching ``Allowance``, if one was found.
        :rtype: Allowance, None
        """

        for allowance in self._allowances:
            if allowance.satisfy_exact_match(args, kwargs):
                return allowance

        for allowance in self._allowances:
            if allowance.satisfy_custom_matcher(args, kwargs):
                return allowance

        for allowance in self._allowances:
            if allowance.satisfy_any_args_match():
                return allowance

    def _find_matching_double(self, args, kwargs):
        """Returns the first matching expectation or allowance.

        Returns the first allowance or expectation that matches the ones declared. Tries one
        with specific arguments first, then falls back to an expectation that allows arbitrary
        arguments.

        :return: The matching ``Allowance`` or ``Expectation``, if one was found.
        :rtype: Allowance, Expectation, None
        """

        expectation = self._find_matching_expectation(args, kwargs)

        if expectation:
            return expectation

        allowance = self._find_matching_allowance(args, kwargs)

        if allowance:
            return allowance

    def _find_matching_expectation(self, args, kwargs):
        """Return a matching expectation.

        Returns the first expectation that matches the ones declared. Tries one with specific
        arguments first, then falls back to an expectation that allows arbitrary arguments.

        :return: The matching ``Expectation``, if one was found.
        :rtype: Expectation, None
        """

        for expectation in self._expectations:
            if expectation.satisfy_exact_match(args, kwargs):
                return expectation

        for expectation in self._expectations:
            if expectation.satisfy_custom_matcher(args, kwargs):
                return expectation

        for expectation in self._expectations:
            if expectation.satisfy_any_args_match():
                return expectation

    def _verify_method(self):
        """Verify that a method may be doubled.

        Verifies that the target object has a method matching the name the user is attempting to
        double.

        :raise: ``VerifyingDoubleError`` if no matching method is found.
        """

        class_level = self._target.is_class_or_module()

        verify_method(self._target, self._method_name, class_level=class_level)



import pytest

from dobles.lifecycle import teardown, verify


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_call(item):
    outcome = yield

    try:
        outcome.get_result()
        try:
            verify()
        except Exception as e:
            outcome.force_exception(e)
    finally:
        teardown()



from functools import wraps
from inspect import isbuiltin
from typing import Set

from dobles.allowance import build_argument_repr_string
from dobles.exceptions import UnallowedMethodCallError
from dobles.proxy_property import ProxyProperty


def double_name(name):
    return "double_of_" + name


_ATTR_METHODS: Set[str] = {
    "__call__",
    "__wrapped__",
    "__aenter__",
    "__aexit__",
    "__enter__",
    "__exit__",
}


def _restore__new__(target, original_method):
    """Restore __new__ to original_method on the target

    Python 3 does some magic to verify no arguments are sent to __new__ if it
    is the builtin version, to work in python 3 we must handle this:

        1) If original_method is the builtin version of __new__, wrap the
        builtin __new__ to ensure that no arguments are passed in.

        2) If original_method is a custom method treat it the same as we would
        in python2

    :param class target: The class to restore __new__ on
    :param func original_method: The method to set __new__ to
    """
    if isbuiltin(original_method):

        @wraps(original_method)
        def _new(cls, *args, **kwargs):
            return original_method(cls)

        target.__new__ = _new
    else:
        target.__new__ = original_method


class ProxyMethod(object):
    """
    The object that replaces the real value of a doubled method. Responsible for hijacking the
    target object, finding matching expectations and returning their values when called, and
    restoring the original value to the hijacked object during teardown.
    """

    def __init__(self, target, method_name, find_expectation):
        """
        :param Target target: The object to be hijacked.
        :param str method_name: The name of the method to replace.
        :param function find_expectation: A function to call to look for expectations that match
             any provided arguments.
        """

        self._target = target
        self._method_name = method_name
        self._find_expectation = find_expectation
        self._attr = target.get_attr(method_name)

        self._capture_original_method()
        self._hijack_target()

    def __call__(self, *args, **kwargs):
        """The actual invocation of the doubled method.

        :param tuple args: The arguments the doubled method was called with.
        :param dict kwargs: The keyword arguments the doubled method was called with.
        :return: The return value the doubled method was declared to return.
        :rtype: object
        :raise: ``UnallowedMethodCallError`` if no matching dobles were found.
        """

        expectation = self._find_expectation(args, kwargs)

        if not expectation:
            self._raise_exception(args, kwargs)

        expectation.verify_arguments(args, kwargs)

        return expectation.return_value(*args, **kwargs)

    def __get__(self, instance, owner):
        """Implements the descriptor protocol to allow doubled properties to behave as properties.

        :return: The return value of any matching double in the case of a property, self otherwise.
        :rtype: object, ProxyMethod
        """

        if self._attr.kind == "property":
            return self.__call__()

        return self

    @property
    def __name__(self):
        return self._original_method.__name__

    @property
    def __doc__(self):
        return self._original_method.__doc__

    @property
    def __wrapped__(self):
        return self._original_method

    def restore_original_method(self):
        """Replaces the proxy method on the target object with its original value."""

        if self._target.is_class_or_module():
            setattr(self._target.obj, self._method_name, self._original_method)
            if self._method_name == "__new__":
                _restore__new__(self._target.obj, self._original_method)
            else:
                setattr(self._target.obj, self._method_name, self._original_method)
        elif self._attr.kind == "property":
            setattr(
                self._target.obj.__class__, self._method_name, self._original_method
            )
            del self._target.obj.__dict__[double_name(self._method_name)]
        elif self._attr.kind == "attribute":
            self._target.obj.__dict__[self._method_name] = self._original_method
        else:
            # TODO: Could there ever have been a value here that needs to be restored?
            del self._target.obj.__dict__[self._method_name]

        if self._method_name in _ATTR_METHODS:
            self._target.restore_attr(self._method_name)

    def _capture_original_method(self):
        """Saves a reference to the original value of the method to be doubled."""

        self._original_method = self._attr.object

    def _hijack_target(self):
        """Replaces the target method on the target object with the proxy method."""

        if self._target.is_class_or_module():
            setattr(self._target.obj, self._method_name, self)
        elif self._attr.kind == "property":
            proxy_property = ProxyProperty(
                double_name(self._method_name),
                self._original_method,
            )
            setattr(self._target.obj.__class__, self._method_name, proxy_property)
            self._target.obj.__dict__[double_name(self._method_name)] = self
        else:
            self._target.obj.__dict__[self._method_name] = self

        if self._method_name in _ATTR_METHODS:
            self._target.hijack_attr(self._method_name)

    def _raise_exception(self, args, kwargs):
        """Raises an ``UnallowedMethodCallError`` with a useful message.

        :raise: ``UnallowedMethodCallError``
        """

        error_message = (
            "Received unexpected call to '{}' on {!r}.  The supplied arguments "
            "{} do not match any available allowances."
        )

        raise UnallowedMethodCallError(
            error_message.format(
                self._method_name,
                self._target.obj,
                build_argument_repr_string(args, kwargs),
            )
        )



from dobles.method_double import MethodDouble
from dobles.target import Target


class Proxy(object):
    """
    An intermediate object used to maintain a mapping between target objects and method dobles.
    """

    def __init__(self, obj):
        """
        :param object obj: The object that will be doubled.
        """

        self._target = Target(obj)
        self._method_dobles = {}

    def add_allowance(self, method_name, caller):
        """Adds a new allowance for the given method name.

        :param str method_name: The name of the method to allow.
        :param tuple caller: A tuple indicating where the allowance was added
        :return: The new ``Allowance``.
        :rtype: Allowance
        """

        return self.method_double_for(method_name).add_allowance(caller)

    def add_expectation(self, method_name, caller):
        """Adds a new expectation for the given method name.

        :param str method_name: The name of the method to expect.
        :return: The new ``Expectation``.
        :rtype: Expectation
        """

        return self.method_double_for(method_name).add_expectation(caller)

    def restore_original_object(self):
        """Remove all stubs from an object.

        Removes the proxy methods applied to each method double and replaces them with the original
        values.
        """

        for method_double in self._method_dobles.values():
            method_double.restore_original_method()

    def verify(self):
        """Verifies all expectations on all method dobles.

        :raise: ``MockExpectationError`` on the first expectation that is not satisfied, if any.
        """

        for method_double in self._method_dobles.values():
            method_double.verify()

    def method_double_for(self, method_name):
        """Returns the method double for the provided method name, creating one if necessary.

        :param str method_name: The name of the method to retrieve a method double for.
        :return: The mapped ``MethodDouble``.
        :rtype: MethodDouble
        """

        if method_name not in self._method_dobles:
            self._method_dobles[method_name] = MethodDouble(method_name, self._target)

        return self._method_dobles[method_name]



class ObjectDouble(object):
    """
    A pure double representing the target object.

    ::

        dummy_user = ObjectDouble(user)

    :param object target: The object the newly created ObjectDouble will verify against.
    """

    is_class = False

    def __init__(self, target):
        self._dobles_target = target

    def __repr__(self):
        """Provides a string representation of the double.

        NOTE: Includes the memory address and the name of the object being doubled.

        :return: A string representation of the double.
        :rtype: str
        """

        address = hex(id(self))
        class_name = self.__class__.__name__

        return "<{} of {!r} object at {}>".format(
            class_name, self._dobles_target, address
        )



from dobles.class_double import ClassDouble
from dobles.lifecycle import current_space


def patch_class(target):
    """
    Replace the specified class with a ClassDouble

    :param str target: A string pointing to the target to patch.
    :param obj values: Values to return when new instances are created.
    :return: A ``ClassDouble`` object.
    """
    class_double = ClassDouble(target)
    patch(target, class_double)
    return class_double


def patch(target, value):
    """
    Replace the specified object

    :param str target: A string pointing to the target to patch.
    :param object value: The value to replace the target with.
    :return: A ``Patch`` object.
    """
    patch = current_space().patch_for(target)
    patch.set_value(value)
    return patch



from importlib import import_module

from dobles.exceptions import VerifyingDoubleImportError


def get_module(module_path, full_path):
    """Return the module given its path.

    :param str module_path: The path to the module to import.
    :param str full_path: The full path to the class that will be doubled.
    :return: The module object.
    :rtype: module
    :raise: ``VerifyingDoubleImportError`` if the module can't be imported.
    """

    try:
        return import_module(module_path)
    except ImportError:
        raise VerifyingDoubleImportError(
            "Cannot import object from path: {}.".format(full_path)
        )


def get_path_components(path):
    """Extract the module name and class name out of the fully qualified path to the class.

    :param str path: The full path to the class.
    :return: The module path and the class name.
    :rtype: str, str
    :raise: ``VerifyingDoubleImportError`` if the path is to a top-level module.
    """

    path_segments = path.split(".")
    module_path = ".".join(path_segments[:-1])

    if module_path == "":
        raise VerifyingDoubleImportError("Invalid import path: {}.".format(path))

    class_name = path_segments[-1]

    return module_path, class_name



import inspect

from dobles.class_double import ClassDouble
from dobles.exceptions import ConstructorDoubleError
from dobles.lifecycle import current_space


def expect(target):
    """
    Prepares a target object for a method call expectation (mock). The name of the method to expect
    should be called as a method on the return value of this function::

        expect(foo).bar

    Accessing the ``bar`` attribute will return an ``Expectation`` which provides additional methods
    to configure the mock.

    :param object target: The object that will be mocked.
    :return: An ``ExpectationTarget`` for the target object.
    """

    return ExpectationTarget(target)


def expect_constructor(target):
    """
    Set an expectation on a ``ClassDouble`` constructor

    :param ClassDouble target:  The ClassDouble to set the expectation on.
    :return: an ``Expectation`` for the __new__ method.
    :raise: ``ConstructorDoubleError`` if target is not a ClassDouble.
    """
    if not isinstance(target, ClassDouble):
        raise ConstructorDoubleError(
            "Cannot allow_constructor of {} since it is not a ClassDouble.".format(
                target
            ),
        )

    return expect(target)._dobles__new__


class ExpectationTarget(object):
    """A wrapper around a target object that creates new expectations on attribute access."""

    def __init__(self, target):
        """
        :param object target: The object to wrap.
        """

        self._proxy = current_space().proxy_for(target)

    def __getattribute__(self, attr_name):
        """
        Returns the value of existing attributes, and returns a new expectation for any attribute
        that doesn't yet exist.

        :param str attr_name: The name of the attribute to look up.
        :return: The existing value or a new ``Expectation``.
        :rtype: object, Expectation
        """

        __dict__ = object.__getattribute__(self, "__dict__")

        if __dict__ and attr_name in __dict__:
            return __dict__[attr_name]

        caller = inspect.getframeinfo(inspect.currentframe().f_back)
        return self._proxy.add_expectation(attr_name, caller)



from contextlib import contextmanager
from threading import local

from dobles.space import Space

_thread_local_data = local()


def current_space():
    """An accessor for the current thread's active ``Space``.

    :return: The active ``Space``.
    :rtype: Space
    """

    if not hasattr(_thread_local_data, "current_space"):
        _thread_local_data.current_space = Space()

    return _thread_local_data.current_space


def teardown():
    """Tears down the current dobles environment. Must be called after each test case."""
    if hasattr(_thread_local_data, "current_space"):
        _thread_local_data.current_space.teardown()
        del _thread_local_data.current_space


def clear(*objects_to_clear):
    """Clears allowances/expectations on objects

    :param object objects_to_clear: The objects to remove allowances and
    expectations from.
    """
    if not hasattr(_thread_local_data, "current_space"):
        return

    space = current_space()
    for obj in objects_to_clear:
        space.clear(obj)


def verify():
    """Verify a mock

    Verifies any mocks that have been created during the test run. Must be called after each
    test case, but before teardown.
    """

    if hasattr(_thread_local_data, "current_space"):
        _thread_local_data.current_space.verify()


@contextmanager
def no_builtin_verification():
    """Turn off validation for builtin methods

    While inside this context we will ignore errors raised while verifying the
    arguments of builtins.

    Note: It is impossible to verify the expected arguments of built in functions
    """
    current_space().skip_builtin_verification = True
    yield
    current_space().skip_builtin_verification = False


def ignore_builtin_verification():
    """Check if we ignoring builtin argument verification errors.

    :return: True if we are ignoring errors.
    :rtype: bool
    """
    return not current_space().skip_builtin_verification



from dobles.allowance import Allowance


class Expectation(Allowance):
    """An individual method expectation (mock)."""

    def __init__(self, target, method_name, caller):
        """
        :param Target target: The object owning the method to mock.
        :param str method_name: The name of the method to mock.
        :param tuple caller: Details of the stack frame where the expectation was made.
        """

        super(Expectation, self).__init__(target, method_name, caller)
        self._is_satisfied = False

    def satisfy_any_args_match(self):
        """
        Returns a boolean indicating whether or not the mock will accept arbitrary arguments.
        This will be true unless the user has specified otherwise using ``with_args`` or
        ``with_no_args``.

        :return: Whether or not the mock accepts arbitrary arguments.
        :rtype: bool
        """

        is_match = super(Expectation, self).satisfy_any_args_match()

        if is_match:
            self._satisfy()

        return is_match

    def satisfy_exact_match(self, args, kwargs):
        """
        Returns a boolean indicating whether or not the mock will accept the provided arguments.

        :return: Whether or not the mock accepts the provided arguments.
        :rtype: bool
        """

        is_match = super(Expectation, self).satisfy_exact_match(args, kwargs)

        if is_match:
            self._satisfy()

        return is_match

    def satisfy_custom_matcher(self, args, kwargs):
        """Returns a boolean indicating whether or not the mock will accept the provided arguments.

        :param tuple args: A tuple of position args
        :param dict kwargs: A dictionary of keyword args
        :return: Whether or not the mock accepts the provided arguments.
        :rtype: bool
        """

        is_match = super(Expectation, self).satisfy_custom_matcher(args, kwargs)

        if is_match:
            self._satisfy()

        return is_match

    def _satisfy(self):
        """Marks the mock as satisfied."""

        self._is_satisfied = True

    def raise_failure_exception(self):
        """
        Raises a ``MockExpectationError`` with a useful message.

        :raise: ``MockExpectationError``
        """

        super(Expectation, self).raise_failure_exception("Expected")

    def is_satisfied(self):
        """
        Returns a boolean indicating whether or not the double has been satisfied. Stubs are
        always satisfied, but mocks are only satisfied if they've been called as was declared,
        or if call is expected not to happen.

        :return: Whether or not the double is satisfied.
        :rtype: bool
        """

        return self._call_counter.has_correct_call_count() and (
            self._call_counter.never() or self._is_satisfied
        )



from inspect import isbuiltin, isfunction, ismethod, signature

from dobles.exceptions import (
    VerifyingBuiltinDoubleArgumentError,
    VerifyingDoubleArgumentError,
    VerifyingDoubleError,
)

ACCEPTS_ARGS = (list, tuple, set)
ACCEPTS_KWARGS = (dict,)

SELF_OR_CLASS = "SELF_OR_CLASS"


def _is_python_function(func):
    if ismethod(func):
        func = func.__func__
    return isfunction(func)


def is_callable(obj):
    if isfunction(obj):
        return True
    return hasattr(obj, "__call__")


def _verify_arguments_of_dobles__new__(target, args, kwargs):
    """Verify arg/kwargs against a class's __init__

    :param class target: The class to verify against.
    :param tuple args:  Positional arguments.
    :params dict kwargs: Keyword arguments.
    """
    if not _is_python_function(target.doubled_obj.__init__):
        class_ = target.doubled_obj
        if args and not kwargs and issubclass(class_, ACCEPTS_ARGS):
            return True
        elif kwargs and not args and issubclass(class_, ACCEPTS_KWARGS):
            return True
        elif args or kwargs:
            given_args_count = 1 + len(args) + len(kwargs)
            raise VerifyingDoubleArgumentError(
                "__init__() takes exactly 1 arguments ({} given)".format(
                    given_args_count,
                )
            )
        return

    return _verify_arguments(
        target.doubled_obj.__init__,
        "__init__",
        [SELF_OR_CLASS] + list(args),
        kwargs,
    )


def _raise_dobles_error_from_index_error(method_name):
    # Work Around for http://bugs.python.org/issue20817
    raise VerifyingDoubleArgumentError(
        "{method}() missing 3 or more arguments.".format(method=method_name)
    )


def verify_method(target, method_name, class_level=False):
    """Verifies that the provided method exists on the target object.

    :param Target target: A ``Target`` object containing the object with the method to double.
    :param str method_name: The name of the method to double.
    :raise: ``VerifyingDoubleError`` if the attribute doesn't exist, if it's not a callable object,
        and in the case where the target is a class, that the attribute isn't an instance method.
    """

    attr = target.get_attr(method_name)

    if not attr:
        raise VerifyingDoubleError(method_name, target.doubled_obj).no_matching_method()

    if (
        attr.kind == "data"
        and not isbuiltin(attr.object)
        and not is_callable(attr.object)
    ):
        raise VerifyingDoubleError(method_name, target.doubled_obj).not_callable()

    if class_level and attr.kind == "method" and method_name != "__new__":
        raise VerifyingDoubleError(method_name, target.doubled_obj).requires_instance()


def verify_arguments(target, method_name, args, kwargs):
    """Verifies that the provided arguments match the signature of the provided method.

    :param Target target: A ``Target`` object containing the object with the method to double.
    :param str method_name: The name of the method to double.
    :param tuple args: The positional arguments the method should be called with.
    :param dict kwargs: The keyword arguments the method should be called with.
    :raise: ``VerifyingDoubleError`` if the provided arguments do not match the signature.
    """

    if method_name == "_dobles__new__":
        return _verify_arguments_of_dobles__new__(target, args, kwargs)

    attr = target.get_attr(method_name)
    method = attr.object

    if attr.kind in (
        "data",
        "attribute",
        "toplevel",
        "class method",
        "static method",
    ):
        try:
            if isinstance(method, (classmethod, staticmethod)):
                method = method.__get__(None, (attr.defining_class))
        except AttributeError:
            method = method.__call__
    elif attr.kind == "property":
        if args or kwargs:
            raise VerifyingDoubleArgumentError("Properties do not accept arguments.")
        return
    else:
        args = [SELF_OR_CLASS] + list(args)

    _verify_arguments(method, method_name, args, kwargs)


def _verify_arguments(method, method_name, args, kwargs):
    try:
        signature(method).bind(*args, **kwargs)
    except ValueError as e:
        raise VerifyingBuiltinDoubleArgumentError(str(e))
    except TypeError as e:
        raise VerifyingDoubleArgumentError(str(e))



from __future__ import absolute_import

import unittest

from dobles.lifecycle import teardown, verify


def wrap_test(test_func):
    def wrapper():
        test_func()
        verify()

    return wrapper


class TestCase(unittest.TestCase):
    def __init__(self, methodName="runTest"):
        super(TestCase, self).__init__(methodName)

        setattr(
            self, self._testMethodName, wrap_test(getattr(self, self._testMethodName))
        )

    def setUp(self):
        self.addCleanup(teardown)



class ProxyProperty(property):
    def __init__(self, name, original):
        """
        :param str name: name of the doubled property
        :param property original: the original property
        """
        self._name = name
        self._original = original

    def __get__(self, obj, objtype=None):
        if self._name in obj.__dict__:
            return obj.__dict__[self._name].__get__(obj, objtype)
        return self._original.__get__(obj, objtype)



class EmptyClass:
    pass


class ArbitraryCallable:
    def __init__(self, value):
        self.value = value

    def __call__(self):
        return self.value


class ArbitraryAsyncCallable:
    def __init__(self, value):
        self.value = value

    async def __call__(self):
        return self.value


async def async_function(*args, **kwargs):
    return "dummy result"


class AsyncUser:
    """An importable dummy class used for testing purposes."""

    class_attribute = "foo"
    callable_class_attribute = classmethod(async_function)
    arbitrary_callable = ArbitraryAsyncCallable("ArbitraryCallable Value")

    def __init__(self, name, age):
        self.name = name
        self.age = age
        self.callable_instance_attribute = async_function

    @staticmethod
    async def static_method(arg):
        return "static_method return value: {}".format(arg)

    @classmethod
    async def class_method(cls, arg):
        return "class_method return value: {}".format(arg)

    async def get_name(self):
        return self.name

    async def instance_method(self):
        return "instance_method return value"

    async def method_with_varargs(self, *args):
        return "method_with_varargs return value"

    async def method_with_default_args(self, foo, bar="baz"):
        return "method_with_default_args return value"

    async def method_with_varkwargs(self, **kwargs):
        return "method_with_varkwargs return value"

    async def method_with_positional_arguments(self, foo):
        return "method_with_positional_arguments return value"

    async def method_with_doc(self):
        """A basic method of User to illustrate existence of a docstring"""
        return

    async def __call__(self, *args):
        return "user was called"

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass


class User:
    """An importable dummy class used for testing purposes."""

    class_attribute = "foo"
    callable_class_attribute = classmethod(lambda cls: "dummy result")
    arbitrary_callable = ArbitraryCallable("ArbitraryCallable Value")

    def __init__(self, name, age):
        self.name = name
        self.age = age
        self.callable_instance_attribute = lambda: "dummy result"

    @staticmethod
    def static_method(arg):
        return "static_method return value: {}".format(arg)

    @classmethod
    def class_method(cls, arg):
        return "class_method return value: {}".format(arg)

    def get_name(self):
        return self.name

    def instance_method(self):
        return "instance_method return value"

    def method_with_varargs(self, *args):
        return "method_with_varargs return value"

    def method_with_default_args(self, foo, bar="baz"):
        return "method_with_default_args return value"

    def method_with_varkwargs(self, **kwargs):
        return "method_with_varkwargs return value"

    def method_with_positional_arguments(self, foo):
        return "method_with_positional_arguments return value"

    def method_with_doc(self):
        """A basic method of User to illustrate existence of a docstring"""
        return

    @property
    def some_property(self):
        return "some_property return value"

    def __call__(self, *args):
        return "user was called"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class UserWithCustomNew(User):
    def __new__(cls, name, age):
        instance = User.__new__(cls)
        instance.name_set_in__new__ = name
        return instance


def top_level_function(arg1, arg2="default"):
    return "{arg1} -- {arg2}".format(arg1=arg1, arg2=arg2)


def top_level_function_that_creates_an_instance():
    return User("Bob Barker", 100), User("Bob Barker", 100)


async def async_top_level_function(arg1, arg2="default"):
    return top_level_function(arg1, arg2)


async def async_top_level_function_that_creates_an_instance():
    return AsyncUser("Bob Barker", 100), AsyncUser("Bob Barker", 100)


class ClassWithGetAttr(object):
    def __init__(self):
        self.attr = "attr"

    def method(self):
        return "method"

    def __getattr__(self, name):
        return "attr {name}".format(name=name)


class Callable(object):
    def __call__(self, arg1):
        return arg1


class AsyncCallable(object):
    async def __call__(self, arg1):
        return arg1


def return_callable(func):
    return Callable()


def return_async_callable(func):
    return AsyncCallable()


def decorate_me(func):
    def decorated(arg1):
        return "{arg1} decorated".format(arg1=arg1)

    return decorated


@return_callable
def decorated_function_callable(arg1):
    return arg1


@decorate_me
def decorated_function(arg1):
    return arg1


@return_async_callable
def async_decorated_function_callable(arg1):
    return arg1


callable_variable = Callable()
async_callable_variable = AsyncCallable()

class_method = User.class_method
instance_method = User("Bob", 25).get_name

async_class_method = AsyncUser.class_method
async_instance_method = AsyncUser("Bob", 25).get_name


class NeverEquals(object):
    def __eq__(self, other):
        return False


class AlwaysEquals(object):
    def __eq__(self, other):
        return True


class DictSubClass(dict):
    pass


class ListSubClass(list):
    pass


class SetSubClass(set):
    pass


class TupleSubClass(tuple):
    pass



from collections import namedtuple
from inspect import (
    classify_class_attrs,
    getmembers,
    isclass,
    iscoroutinefunction,
    ismodule,
)
from typing import Any

from dobles.object_double import ObjectDouble
from dobles.verification import is_callable

Attribute = namedtuple("Attribute", ["object", "kind", "defining_class"])


def _proxy_class_method_to_instance(original, name):
    def func(instance, *args, **kwargs):
        if name in instance.__dict__:
            return instance.__dict__[name](*args, **kwargs)
        return original(instance, *args, **kwargs)

    func._dobles_target_method = original
    return func


class Target(object):
    """
    A wrapper around an object that owns methods to be doubled. Provides additional introspection
    such as the class and the kind (method, class, property, etc.)
    """

    def __init__(self, obj):
        """
        :param object obj: The real target object.
        """

        self.obj = obj
        self.doubled_obj = self._determine_doubled_obj()
        self.doubled_obj_type = self._determine_doubled_obj_type()
        self.attrs = self._generate_attrs()

    def is_class_or_module(self):
        """Determines if the object is a class or a module

        :return: True if the object is a class or a module, False otherwise.
        :rtype: bool
        """

        if isinstance(self.obj, ObjectDouble):
            return self.obj.is_class

        return isclass(self.doubled_obj) or ismodule(self.doubled_obj)

    def _determine_doubled_obj(self):
        """Return the target object.

        Returns the object that should be treated as the target object. For partial dobles, this
        will be the same as ``self.obj``, but for pure dobles, it's pulled from the special
        ``_dobles_target`` attribute.

        :return: The object to be doubled.
        :rtype: object
        """

        if isinstance(self.obj, ObjectDouble):
            return self.obj._dobles_target
        else:
            return self.obj

    def _determine_doubled_obj_type(self):
        """Returns the type (class) of the target object.

        :return: The type (class) of the target.
        :rtype: type, classobj
        """

        if isclass(self.doubled_obj) or ismodule(self.doubled_obj):
            return self.doubled_obj

        return self.doubled_obj.__class__

    def _generate_attrs(self):
        """Get detailed info about target object.

        Uses ``inspect.classify_class_attrs`` to get several important details about each attribute
        on the target object.

        :return: The attribute details dict.
        :rtype: dict
        """
        attrs = {}

        if ismodule(self.doubled_obj):
            for name, func in getmembers(self.doubled_obj, is_callable):
                attrs[name] = Attribute(func, "toplevel", self.doubled_obj)
        else:
            for attr in classify_class_attrs(self.doubled_obj_type):
                attrs[attr.name] = attr

        return attrs

    def hijack_attr(self, attr_name):
        """Hijack an attribute on the target object.

        Updates the underlying class and delegating the call to the instance.
        This allows specially-handled attributes like __call__, __enter__,
        and __exit__ to be mocked on a per-instance basis.

        :param str attr_name: the name of the attribute to hijack
        """
        if not self._original_attr(attr_name):
            setattr(
                self.obj.__class__,
                attr_name,
                _proxy_class_method_to_instance(
                    getattr(self.obj.__class__, attr_name, None), attr_name
                ),
            )

    def restore_attr(self, attr_name):
        """Restore an attribute back onto the target object.

        :param str attr_name: the name of the attribute to restore
        """
        original_attr = self._original_attr(attr_name)
        if self._original_attr(attr_name):
            setattr(self.obj.__class__, attr_name, original_attr)

    def _original_attr(self, attr_name):
        """Return the original attribute off of the proxy on the target object.

        :param str attr_name: the name of the original attribute to return
        :return: Func or None.
        :rtype: func
        """
        try:
            return getattr(
                getattr(self.obj.__class__, attr_name), "_dobles_target_method", None
            )
        except AttributeError:
            return None

    def get_callable_attr(self, attr_name):
        """Used to double methods added to an object after creation

        :param str attr_name: the name of the original attribute to return
        :return: Attribute or None.
        :rtype: func
        """

        if not hasattr(self.doubled_obj, attr_name):
            return None

        func = getattr(self.doubled_obj, attr_name)
        if not is_callable(func):
            return None

        attr = Attribute(
            func,
            "attribute",
            self.doubled_obj if self.is_class_or_module() else self.doubled_obj_type,
        )
        self.attrs[attr_name] = attr
        return attr

    def get_attr(self, method_name):
        """Get attribute from the target object"""
        return self.attrs.get(method_name) or self.get_callable_attr(method_name)

    def is_attr_async(self, name: str) -> bool:
        attr: Any = self.get_attr(name)
        object: Any = attr.object
        object = getattr(object, "_dobles_target_method", object)

        if iscoroutinefunction(object):
            return True

        if isinstance(object, classmethod):
            object = object.__get__(None, (attr.defining_class))

        else:
            for i in ("__call__", "__wrapped__"):
                if hasattr(object, i):
                    object = getattr(object, i)
                    break

        return iscoroutinefunction(object)



from dobles.patch import Patch
from dobles.proxy import Proxy


class Space(object):
    """
    A container object for all the dobles created during the execution of a test case. Maintains
    a one-to-one mapping of target objects and ``Proxy`` objects. Maintained by the ``lifecycle``
    module and not intended to be used directly by other objects.
    """

    def __init__(self):
        self._proxies = {}
        self._patches = {}
        self._is_verified = False
        self.skip_builtin_verification = False

    def patch_for(self, path):
        """Returns the ``Patch`` for the target path, creating it if necessary.

        :param str path: The absolute module path to the target.
        :return: The mapped ``Patch``.
        :rtype: Patch
        """

        if path not in self._patches:
            self._patches[path] = Patch(path)

        return self._patches[path]

    def proxy_for(self, obj):
        """Returns the ``Proxy`` for the target object, creating it if necessary.

        :param object obj: The object that will be doubled.
        :return: The mapped ``Proxy``.
        :rtype: Proxy
        """

        obj_id = id(obj)

        if obj_id not in self._proxies:
            self._proxies[obj_id] = Proxy(obj)

        return self._proxies[obj_id]

    def teardown(self):
        """Restores all doubled objects to their original state."""

        for proxy in self._proxies.values():
            proxy.restore_original_object()

        for patch in self._patches.values():
            patch.restore_original_object()

    def clear(self, obj):
        """Clear allowances/expectations set on an object.

        :param object obj: The object to clear.
        """
        self.proxy_for(obj).restore_original_object()
        del self._proxies[id(obj)]

    def verify(self):
        """Verifies expectations on all doubled objects.

        :raise: ``MockExpectationError`` on the first expectation that is not satisfied, if any.
        """

        if self._is_verified:
            return

        for proxy in self._proxies.values():
            proxy.verify()

        self._is_verified = True



class MockExpectationError(AssertionError):
    """An exception raised when a mock fails verification."""

    pass


class UnallowedMethodCallError(AssertionError):
    """An exception raised when an unallowed method call is made on a double."""

    pass


class VerifyingDoubleError(AssertionError):
    """
    An exception raised when attempting to double a method that does not exist on the real object.
    """

    def __init__(self, method_name, doubled_obj):
        """
        :param str method_name: The name of the method to double.
        :param object doubled_obj: The real object being doubled.
        """

        self._method_name = method_name
        self._doubled_obj = doubled_obj
        self.args = (method_name, doubled_obj)
        self.message = "Cannot double method '{}' of '{}'."

    def no_matching_method(self):
        self.message = "Cannot double method '{}' because {} does not implement it."

        return self

    def not_callable(self):
        self.message = (
            "Cannot double method '{}' because it is not a callable attribute on {}."
        )

        return self

    def requires_instance(self):
        self.message = (
            "Cannot double method '{}' because it is not callable directly on {}."
        )

        return self

    def __str__(self):
        return self.message.format(self._method_name, self._doubled_obj)


class VerifyingDoubleArgumentError(AssertionError):
    """
    An exception raised when attempting to double a method with arguments that do not match the
    signature of the real method.
    """

    pass


class VerifyingBuiltinDoubleArgumentError(VerifyingDoubleArgumentError):
    """
    An exception raised when attempting to validate arguments of a builtin.
    """

    pass


class VerifyingDoubleImportError(AssertionError):
    """
    An exception raised when attempting to create a verifying double from an invalid module path.
    """


class ConstructorDoubleError(AssertionError):
    """
    An exception raised when attempting to double the constructor of a non ClassDouble.
    """



from dobles.exceptions import VerifyingDoubleError
from dobles.utils import get_module, get_path_components


class Patch(object):
    """
    A wrapper around an object that has been ``patched``
    """

    def __init__(self, target):
        """
        :param str path: The absolute module path to the class.
        """
        module_path, self._name = get_path_components(target)
        self.target = get_module(module_path, target)
        self._capture_original_object()
        self.set_value(None)

    def _capture_original_object(self):
        """Capture the original python object."""
        try:
            self._dobles_target = getattr(self.target, self._name)
        except AttributeError:
            raise VerifyingDoubleError(self.target, self._name)

    def set_value(self, value):
        """Set the value of the target.

        :param obj value: The value to set.
        """
        self._value = value
        setattr(self.target, self._name, value)

    def restore_original_object(self):
        """Restore the target to it's original value."""
        self.set_value(self._dobles_target)



__version__ = "2.0.0"

from dobles.class_double import ClassDouble  # noqa
from dobles.instance_double import InstanceDouble  # noqa
from dobles.lifecycle import clear, no_builtin_verification, teardown, verify  # noqa
from dobles.object_double import ObjectDouble  # noqa
from dobles.targets.allowance_target import allow, allow_constructor  # noqa
from dobles.targets.expectation_target import expect, expect_constructor  # noqa
from dobles.targets.patch_target import patch, patch_class  # noqa
